{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "2fc57809",
   "metadata": {},
   "source": [
    "# Outlook\n",
    "In this notebook, using BBRL, we code a simple version of the DQN algorithm\n",
    "without a replay buffer nor a target network so as to better understand the\n",
    "inner mechanisms.\n",
    "\n",
    "To understand this code, you need to know more about \n",
    "[the BBRL interaction model](https://github.com/osigaud/bbrl/blob/master/docs/overview.md)\n",
    "Then you should run [a didactical example](https://github.com/osigaud/bbrl/blob/master/docs/notebooks/02-multi_env_noautoreset.student.ipynb)\n",
    "to see how agents interact in BBRL when autoreset=False.\n",
    "\n",
    "The DQN algorithm is explained in [this\n",
    "video](https://www.youtube.com/watch?v=CXwvOMJujZk) and you can also read [the\n",
    "corresponding slides](http://pages.isir.upmc.fr/~sigaud/teach/dqn.pdf)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "509112e7",
   "metadata": {},
   "source": [
    "## Installation and Imports\n",
    "\n",
    "### Installation\n",
    "\n",
    "The BBRL library is [here](https://github.com/osigaud/bbrl).\n",
    "\n",
    "We use OmegaConf to that makes it possible that by just defining the `def\n",
    "run_dqn(cfg):` function and then executing a long `params = {...}` variable at\n",
    "the bottom of this colab, the code is run with the parameters without calling\n",
    "an explicit main.\n",
    "\n",
    "More precisely, the code is run by calling\n",
    "\n",
    "`config=OmegaConf.create(params)`\n",
    "\n",
    "`run_dqn(config)`\n",
    "\n",
    "at the very bottom of the colab, after starting tensorboard.\n",
    "\n",
    "Below, we import standard python packages, pytorch packages and gymnasium\n",
    "environments."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "424cea46",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Installs the necessary Python and system libraries\n",
    "try:\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "except ModuleNotFoundError as e:\n",
    "    get_ipython().run_line_magic(\"pip\", \"install easypip\")\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "\n",
    "easyinstall(\"bbrl>=0.2.2\")\n",
    "easyinstall(\"swig\")\n",
    "easyinstall(\"bbrl_gymnasium>=0.2.0\")\n",
    "easyinstall(\"bbrl_gymnasium[box2d]\")\n",
    "easyinstall(\"bbrl_gymnasium[classic_control]\")\n",
    "easyinstall(\"tensorboard\")\n",
    "easyinstall(\"moviepy\")\n",
    "easyinstall(\"box2d-kengz\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0c7ea5ad",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import os\n",
    "import sys\n",
    "from pathlib import Path\n",
    "import math\n",
    "\n",
    "from moviepy.editor import ipython_display as video_display\n",
    "import time\n",
    "from tqdm.auto import tqdm\n",
    "from typing import Tuple, Optional\n",
    "from functools import partial\n",
    "\n",
    "from omegaconf import OmegaConf\n",
    "import torch\n",
    "import bbrl_gymnasium\n",
    "\n",
    "import copy\n",
    "from abc import abstractmethod, ABC\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from time import strftime\n",
    "\n",
    "OmegaConf.register_new_resolver(\n",
    "    \"current_time\", lambda: strftime(\"%Y%m%d-%H%M%S\"), replace=True\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "86f04e5f",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Imports all the necessary classes and functions from BBRL\n",
    "from bbrl.agents.agent import Agent\n",
    "from bbrl import get_arguments, get_class, instantiate_class\n",
    "\n",
    "# The workspace is the main class in BBRL, this is where all data is collected and stored\n",
    "from bbrl.workspace import Workspace\n",
    "\n",
    "# Agents(agent1,agent2,agent3,...) executes the different agents the one after the other\n",
    "# TemporalAgent(agent) executes an agent over multiple timesteps in the workspace,\n",
    "# or until a given condition is reached\n",
    "from bbrl.agents import Agents, TemporalAgent\n",
    "\n",
    "# ParallelGymAgent is an agent able to execute a batch of gymnasium environments\n",
    "# with auto-resetting. These agents produce multiple variables in the workspace:\n",
    "# ’env/env_obs’, ’env/reward’, ’env/timestep’, ’env/terminated’,\n",
    "# 'env/truncated', 'env/done', ’env/cumulated_reward’, ...\n",
    "#\n",
    "# When called at timestep t=0, the environments are automatically reset. At\n",
    "# timestep t>0, these agents will read the ’action’ variable in the workspace at\n",
    "# time t − 1\n",
    "from bbrl.agents.gymnasium import GymAgent, ParallelGymAgent, make_env, record_video\n",
    "\n",
    "# Replay buffers are useful to store past transitions when training\n",
    "from bbrl.utils.replay_buffer import ReplayBuffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a3f545dc",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Utility function for launching tensorboard\n",
    "# For Colab - otherwise, it is easier and better to launch tensorboard from\n",
    "# the terminal\n",
    "def setup_tensorboard(path):\n",
    "    path = Path(path)\n",
    "    answer = \"\"\n",
    "    if is_notebook():\n",
    "        if get_ipython().__class__.__module__ == \"google.colab._shell\":\n",
    "            answer = \"y\"\n",
    "        while answer not in [\"y\", \"n\"]:\n",
    "            answer = input(\n",
    "                f\"Do you want to launch tensorboard in this notebook [y/n] \"\n",
    "            ).lower()\n",
    "\n",
    "    if answer == \"y\":\n",
    "        get_ipython().run_line_magic(\"load_ext\", \"tensorboard\")\n",
    "        get_ipython().run_line_magic(\"tensorboard\", f\"--logdir {path.absolute()}\")\n",
    "    else:\n",
    "        import sys\n",
    "        import os\n",
    "        import os.path as osp\n",
    "\n",
    "        print(\n",
    "            f\"Launch tensorboard from the shell:\\n{osp.dirname(sys.executable)}/tensorboard --logdir={path.absolute()}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "001c40d0",
   "metadata": {},
   "source": [
    "## Definition of agents\n",
    "\n",
    "The [DQN](https://daiwk.github.io/assets/dqn.pdf) algorithm is a critic only algorithm.\n",
    "Thus we just need a Critic agent (which is also used to output actions) and an Environment agent.\n",
    "\n",
    "### The critic agent\n",
    "\n",
    "The critic agent is an instance of the `DiscreteQAgent` class.\n",
    "We first build a deterministic neural network that takes the state as input (so it has one input neuron per state variable)\n",
    "and that outputs the Q-value of each action in that state (so it has one output neuron per action)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4847e610",
   "metadata": {},
   "source": [
    "The function below builds a multi-layer perceptron where the size of each layer is given in the `size` list.\n",
    "We also specify the activation function of neurons at each layer and optionally a different activation function for the final layer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "154cb37d",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch.nn as nn\n",
    "def build_mlp(sizes, activation, output_activation=nn.Identity()):\n",
    "    \"\"\"Helper function to build a multi-layer perceptron (function from $\\mathbb R^n$ to $\\mathbb R^p$)\n",
    "    \n",
    "    Args:\n",
    "        sizes (List[int]): the number of neurons at each layer\n",
    "        activation (nn.Module): a PyTorch activation function (after each layer but the last)\n",
    "        output_activation (nn.Module): a PyTorch activation function (last layer)\n",
    "    \"\"\"\n",
    "    layers = []\n",
    "    for j in range(len(sizes) - 1):\n",
    "        act = activation if j < len(sizes) - 2 else output_activation\n",
    "        layers += [nn.Linear(sizes[j], sizes[j + 1]), act]\n",
    "    return nn.Sequential(*layers)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bba74506",
   "metadata": {},
   "source": [
    "As any BBRL agent, the DiscreteQAgent has a `forward()` function that takes a time state as input.\n",
    "This `forward()` function outputs the Q-values of all actions at the corresponding time step.\n",
    "Additionally, if the critic is used to choose an action, it also outputs the chosen action at the same time step."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "101c1fc1",
   "metadata": {},
   "outputs": [],
   "source": [
    "class DiscreteQAgent(Agent):\n",
    "    \"\"\"BBRL agent (discrete actions) based on a MLP\"\"\"\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim):\n",
    "        super().__init__()\n",
    "        self.model = build_mlp(\n",
    "            [state_dim] + list(hidden_layers) + [action_dim], activation=nn.ReLU()\n",
    "        )\n",
    "\n",
    "    def forward(self, t: int, **kwargs):\n",
    "        \"\"\"An Agent can use self.workspace\"\"\"\n",
    "\n",
    "        # Retrieves the observation from the environment at time t\n",
    "        obs = self.get((\"env/env_obs\", t))\n",
    "\n",
    "        # Computes the critic (Q) values for the observation\n",
    "        q_values = self.model(obs)\n",
    "\n",
    "        # ... and sets the q-values (one for each possible action)\n",
    "        self.set((\"q_values\", t), q_values)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a1d6f089",
   "metadata": {},
   "source": [
    "#### Greedily choosing the action\n",
    "\n",
    "The ArgmaxActionSelector is in charge of choosing the action whose Q-value is the highest given the Q-values of all actions.\n",
    "We may use it when we do not want to explore. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "225417bb",
   "metadata": {},
   "outputs": [],
   "source": [
    "class ArgmaxActionSelector(Agent):\n",
    "    \"\"\"BBRL agent that selects the best action based on Q(s,a)\"\"\"\n",
    "    def forward(self, t: int, **kwargs):\n",
    "        q_values = self.get((\"q_values\", t))\n",
    "        action = q_values.argmax(1)\n",
    "        self.set((\"action\", t), action)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "881f44e6",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Creating an Exploration method\n",
    "\n",
    "As Q-learning, DQN needs some exploration to prevent too early convergence.\n",
    "Here we use the simple $\\epsilon$-greedy exploration method.\n",
    "It is implemented as an agent which chooses an action based on the Q-values."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ea994709",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class EGreedyActionSelector(Agent):\n",
    "    def __init__(self, epsilon):\n",
    "        super().__init__()\n",
    "        self.epsilon = epsilon\n",
    "\n",
    "    def forward(self, t: int, **kwargs):\n",
    "        # Retrieves the q values \n",
    "        # (matrix nb. of episodes x nb. of actions)\n",
    "        q_values = self.get((\"q_values\", t))\n",
    "        size, nb_actions = q_values.size()\n",
    "\n",
    "        # Flag \n",
    "        is_random = torch.rand(size).lt(self.epsilon).float()\n",
    "        random_action = torch.randint(low=0, high=nb_actions, size=(size,))\n",
    "        max_action = q_values.max(1)[1]\n",
    "\n",
    "        # Choose the action based on the is_random flag\n",
    "        action = is_random * random_action + (1 - is_random) * max_action\n",
    "\n",
    "        # Sets the action at time t\n",
    "        self.set((\"action\", t), action.long())"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "42c18889",
   "metadata": {},
   "source": [
    "### The Logger class\n",
    "\n",
    "The logger is in charge of collecting statistics during the training\n",
    "process.\n",
    "\n",
    "Having logging provided under the hood is one of the features allowing you\n",
    "to save time when using RL libraries like BBRL.\n",
    "\n",
    "In these notebooks, the logger is defined as `bbrl.utils.logger.TFLogger` so as\n",
    "to use a tensorboard visualisation (see the parameters part `params = { \"logger\":{ ...` below).\n",
    "\n",
    "Note that the BBRL Logger is also saving the log in a readable format such\n",
    "that you can use `Logger.read_directories(...)` to read multiple logs, create\n",
    "a dataframe, and analyze many experiments afterward in a notebook for\n",
    "instance. The code for the different kinds of loggers is available in the\n",
    "[bbrl/utils/logger.py](https://github.com/osigaud/bbrl/blob/master/src/bbrl/utils/logger.py)\n",
    "file.\n",
    "\n",
    "`instantiate_class` is an inner BBRL mechanism. The\n",
    "`instantiate_class`function is available in the\n",
    "[`bbrl/__init__.py`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/__init__.py)\n",
    "file."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "251d8534",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "from bbrl import instantiate_class\n",
    "\n",
    "class Logger():\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        self.logger = instantiate_class(cfg.logger)\n",
    "\n",
    "    def add_log(self, log_string, loss, steps):\n",
    "        self.logger.add_scalar(log_string, loss.item(), steps)\n",
    "\n",
    "    # A specific function for RL algorithms having a critic, an actor and an entropy losses\n",
    "    def log_losses(self, critic_loss, entropy_loss, actor_loss, steps):\n",
    "        self.add_log(\"critic_loss\", critic_loss, steps)\n",
    "        self.add_log(\"entropy_loss\", entropy_loss, steps)\n",
    "        self.add_log(\"actor_loss\", actor_loss, steps)\n",
    "\n",
    "    def log_reward_losses(self, rewards, nb_steps):\n",
    "        self.add_log(\"reward/mean\", rewards.mean(), nb_steps)\n",
    "        self.add_log(\"reward/max\", rewards.max(), nb_steps)\n",
    "        self.add_log(\"reward/min\", rewards.min(), nb_steps)\n",
    "        self.add_log(\"reward/median\", rewards.median(), nb_steps)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "26b598a7",
   "metadata": {},
   "source": [
    "## Heart of the algorithm\n",
    "### Computing the critic loss\n",
    "The role of the `compute_critic_loss` function is to implement the Bellman\n",
    "backup rule. In Q-learning, this rule was written:\n",
    "\n",
    "$$Q(s_t,a_t) \\leftarrow Q(s_t,a_t) + \\alpha [ r(s_t,a_t) + \\gamma \\max_a\n",
    "Q(s_{t+1},a) - Q(s_t,a_t)]$$\n",
    "\n",
    "In DQN, the update rule $Q \\leftarrow Q + \\alpha [\\delta] $ is replaced by a\n",
    "gradient descent step over the Q-network. \n",
    "\n",
    "We first compute a target value: $ target = r(s_t,a_t) + \\gamma \\max_a\n",
    "Q(s_{t+1},a)$ from a set of samples.\n",
    "\n",
    "Then we get a TD error $\\delta$ by substracting $Q(s_t,a_t)$ for these samples, \n",
    "and we use the squared TD error as a loss function: $ loss = (target -\n",
    "Q(s_t,a_t))^2$.\n",
    "\n",
    "To implement the above calculation in BBRL, the difficulty is to\n",
    "properly deal with time indexes.\n",
    "\n",
    "The `compute_critic_loss` function receives rewards, q_values and actions as\n",
    "tensors that have been computed over a complete episode.\n",
    "\n",
    "We need to take `reward[1:]`, which means all the rewards except the first one\n",
    "because the reward from $(s_t, a_t)$ is $r_{t+1}$.\n",
    "Similarly, to get $\\max_a Q(s_{t+1}, a)$, we need to ignore the first of the\n",
    "max_q values, using `max_q[1:]`.\n",
    "\n",
    "Do not forget to apply .detach() when computing the values of $\\max_a Q(s_{t+1}, a)$, as **we do not\n",
    "want to apply gradient descent on this $\\max_a Q(s_{t+1}, a)$**, we only\n",
    "apply gradient descent to $Q(s_t, a_t)$ according to this target value.\n",
    "In practice, `x.detach()` detaches a computation graph from a tensor,\n",
    "so it avoids computing a gradient over this tensor.\n",
    "\n",
    "The `must_bootstrap` tensor is used as a trick to deal with terminal states,\n",
    "as explained [here](https://github.com/osigaud/bbrl/blob/master/docs/time_limits.md)\n",
    "In practice, `must_bootstrap` is the logical negation of `terminated`.\n",
    "In the autoreset=False version we use full episodes, thus `must_bootstrap` is\n",
    "always True for all steps but the last one.\n",
    "\n",
    "To compute $Q(s_t,a_t)$ we use the `torch.gather()` function. This function is\n",
    "a little tricky to use, see [this page](https://github.com/osigaud/bbrl/blob/master/docs/using_gather.md)\n",
    "for useful explanations.\n",
    "\n",
    "In particular, the q_vals output that we get is not properly conditioned,\n",
    "hence the need for the `qval[:-1]` (we ignore the last dimension). Finally we\n",
    "just need to compute the difference target - qvals, square it, take the mean\n",
    "and send it back as the loss."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "124c4eec",
   "metadata": {},
   "outputs": [],
   "source": [
    "def compute_critic_loss(cfg, reward: torch.Tensor, must_bootstrap: torch.Tensor, q_values: torch.Tensor, action: torch.LongTensor) -> torch.Tensor:\n",
    "    \"\"\"Compute the temporal difference loss from a dataset to \n",
    "    update a critic\n",
    "\n",
    "    For the tensor dimensions:\n",
    "    \n",
    "    - T = maximum number of time steps\n",
    "    - B = number of episodes run in parallel \n",
    "    - A = action space dimension\n",
    "\n",
    "    :param cfg: The configuration\n",
    "    :param reward: A (T x B) tensor containing the rewards \n",
    "    :param must_bootstrap: a (T x B) tensor containing 0 if the episode is\n",
    "        completed at time $t$ \n",
    "    :param q_values: a (T x B x A) tensor containing the Q-values at each\n",
    "        time step\n",
    "    :param action: a (T x B) long tensor containing the chosen action\n",
    "\n",
    "    :return: The DQN loss\n",
    "    \"\"\"\n",
    "    # We compute the max of Q-values over all actions and detach (so that\n",
    "    # this part of the computation graph is not included in the gradient\n",
    "    # backpropagation)\n",
    "\n",
    "    # Compute the loss\n",
    "\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "    return critic_loss"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "dd0fc0a2",
   "metadata": {},
   "source": [
    "## Main training loop\n",
    "\n",
    "Note that everything about the shared workspace between all the agents is\n",
    "completely hidden under the hood. This results in a gain of productivity, at\n",
    "the expense of having to dig into the BBRL code if you want to understand the\n",
    "details, change the multiprocessing model, etc.\n",
    "\n",
    "The next cell defines a `EpisodicDQN` that deals with various part of the training\n",
    "loop:\n",
    "\n",
    "- `__init__` takes care of initializing the train and evaluation policies"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ad912e79",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl import get_arguments, get_class\n",
    "from itertools import chain\n",
    "\n",
    "def setup_optimizer(cfg_optimizer, *agents):\n",
    "    \"\"\"Setup an optimizer for a list of agents\"\"\"\n",
    "    optimizer_args = get_arguments(cfg_optimizer)\n",
    "    parameters = [agent.parameters() for agent in agents]\n",
    "    optimizer = get_class(cfg_optimizer)(chain(*parameters), **optimizer_args)\n",
    "    return optimizer\n",
    "\n",
    "def copy_parameters(model_a, model_b):\n",
    "    \"\"\"Copy parameters from a model a to model_b\"\"\"\n",
    "    for model_a_p, model_b_p in zip(model_a.parameters(), model_b.parameters()):\n",
    "        model_b_p.data.copy_(model_a_p)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0c76e0c0",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "# Learning environment\n",
    "\n",
    "To setup a common learning environment for RL algorithms, we use the `RLBase`\n",
    "class. This class:\n",
    "\n",
    "1. Initializes the environment (random seed, logger, evaluation environment)\n",
    "2. Defines a `evaluate` method that keeps the best agent so far\n",
    "3. Defines a `visualize_best` method that displays the behavior of the best agent\n",
    "\n",
    "Subclasses need to define `self.train_policy` and `self.eval_policy`, two\n",
    "BBRL agents that respectively choose actions when training and evaluating.\n",
    "\n",
    "The behavior of `RLBase` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `base_dir` defines the directory subpath used when outputing losses during\n",
    "training as well as other outputs (serialized agent, global statistics, etc.)\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `gym_env` defines the gymnasium environment, and in particular\n",
    "`gym_env.env_name` the name of the gymansium environment\n",
    "- `logger` defines what type of logger is used to log the different values\n",
    "associated with learning\n",
    "- `algorithm.eval_interval` defines the number of observed transitions between\n",
    "each evaluation of the agent"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7b069c0a",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from typing import Any\n",
    "import logging\n",
    "from abc import ABC\n",
    "from functools import cached_property\n",
    "\n",
    "class RLBase(ABC):\n",
    "    \"\"\"Base class for Reinforcement learning algorithms\n",
    "    \n",
    "    This class deals with common processing:\n",
    "\n",
    "    - defines the logger, the train and evaluation agents\n",
    "    - defines how to evaluate a policy\n",
    "    \"\"\"\n",
    "\n",
    "    #: The configuration\n",
    "    cfg: Any\n",
    "\n",
    "    #: The evaluation environment deals with the last action, and produces a new\n",
    "    # state of the environment\n",
    "    eval_env: Agent\n",
    "\n",
    "    #: The training policy\n",
    "    train_policy: Agent\n",
    "\n",
    "    #: The evaluation policy (if not defined, uses the training policy)\n",
    "    eval_policy: Agent\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        # Basic initialization\n",
    "        self.cfg = cfg\n",
    "        torch.manual_seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Sets the base directory and logger directory\n",
    "        base_dir = Path(self.cfg.base_dir)\n",
    "        self.base_dir = Path(\"outputs\") / Path(self.cfg.base_dir)\n",
    "        self.base_dir.mkdir(parents=True, exist_ok=True)\n",
    "\n",
    "        # Initialize the logger class\n",
    "        if not hasattr(cfg.logger, \"log_dir\"):\n",
    "            cfg.logger.log_dir = str(Path(\"outputs\") / \"tblogs\" / base_dir)\n",
    "        self.logger = Logger(cfg)\n",
    "\n",
    "        # Subclasses have to define the training and eval policies\n",
    "        self.train_policy = None\n",
    "        self.eval_policy = None\n",
    "\n",
    "        # Sets up the evaluation environment\n",
    "        self.eval_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name), \n",
    "            cfg.algorithm.nb_evals\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Initialize values\n",
    "        self.last_eval_step = 0\n",
    "        self.nb_steps = 0\n",
    "        self.best_policy = None\n",
    "        self.best_reward = -torch.inf\n",
    "\n",
    "        # Records the rewards\n",
    "        self.eval_rewards = []\n",
    "\n",
    "    @cached_property\n",
    "    def train_agent(self):\n",
    "        \"\"\"Returns the training agent\n",
    "\n",
    "        The agent is composed of a policy agent and the training environment.      \n",
    "        This method supposes that `self.train_policy` has been setup\n",
    "        \"\"\"\n",
    "        assert self.train_policy is not None, \"The train_policy property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.train_env, self.train_policy))\n",
    "\n",
    "    @cached_property\n",
    "    def eval_agent(self):\n",
    "        \"\"\"Returns the evaluation agent \n",
    "        \n",
    "        The agent is composed of a policy agent and the evaluation environment\n",
    "        \n",
    "        Uses `self.eval_policy` (or `self.train_policy` if not defined)\n",
    "\n",
    "        \"\"\"\n",
    "        assert self.eval_policy is not None or self.train_policy is not None, \"eval_agent property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.eval_env, self.eval_policy if self.eval_policy is not None else self.train_policy))\n",
    "\n",
    "    def evaluate(self):\n",
    "        \"\"\"Evaluate the current policy `self.eval_policy`\n",
    "        \n",
    "        Evaluation is conducted every `cfg.algorithm.eval_interval` steps, and\n",
    "        we keep a copy of the best agent so far in `self.best_policy`\n",
    "        \n",
    "        Returns True if the current policy is the best so far\n",
    "        \"\"\"\n",
    "        if (self.nb_steps - self.last_eval_step) > self.cfg.algorithm.eval_interval:\n",
    "            self.last_eval_step = self.nb_steps\n",
    "            eval_workspace = Workspace() \n",
    "            self.eval_agent(\n",
    "                eval_workspace,\n",
    "                t=0,\n",
    "                stop_variable=\"env/done\"\n",
    "            )\n",
    "            rewards = eval_workspace[\"env/cumulated_reward\"][-1]\n",
    "            self.logger.log_reward_losses(rewards, self.nb_steps)\n",
    "\n",
    "            if getattr(self.cfg, \"collect_stats\", False):\n",
    "                self.eval_rewards.append(rewards)\n",
    "\n",
    "            rewards_mean = rewards.mean()\n",
    "            if rewards_mean > self.best_reward:\n",
    "                self.best_policy = copy.deepcopy(self.eval_policy)\n",
    "                self.best_reward = rewards_mean\n",
    "                return True\n",
    "\n",
    "    def save_stats(self):\n",
    "        \"\"\"Save reward statistics into `stats.npy`\"\"\"\n",
    "        if getattr(self.cfg, \"collect_stats\", False) and self.eval_rewards:\n",
    "            data = torch.stack(self.eval_rewards, axis=-1) \n",
    "            with (self.base_dir / \"stats.npy\").open(\"wt\") as fp:\n",
    "                np.savetxt(fp, data.numpy())\n",
    "\n",
    "    def visualize_best(self):\n",
    "        \"\"\"Visualize the best agent\"\"\"\n",
    "        env = make_env(self.cfg.gym_env.env_name, render_mode=\"rgb_array\")\n",
    "        path = self.base_dir / \"best_agent.mp4\"\n",
    "        print(f\"Video of best agent recorded in {path}\")\n",
    "        record_video(env, self.best_policy, path)\n",
    "        return video_display(str(path.absolute()))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "251b54e1",
   "metadata": {
    "lines_to_next_cell": 0
   },
   "source": [
    "The `EpisodicAlgo` defines the environment when using episodes. In particular,\n",
    "it defines `self.train_env` which is the environment used for training. As\n",
    "`algorithm.n_envs` are used in parallel, when a episode ends, we don't stop\n",
    "the other episodes. To cater for this:\n",
    "\n",
    "1. the workspace variable `env/done` is set to `True` for all the next time\n",
    "steps\n",
    "2. The variable `env/reward` is set to 0 for all the steps \n",
    "\n",
    "The behavior of `EpisodicAlgo` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `gym_env.env_name` defines the gymnasium environment\n",
    "- `algorithm.n_envs` defines the number of parallel environments\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b2aba9de",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "class EpisodicAlgo(RLBase):\n",
    "    \"\"\"Base class for RL experiments with full episodes\"\"\"\n",
    "    def __init__(self, cfg, autoreset=False):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        self.train_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name, autoreset=autoreset), \n",
    "            cfg.algorithm.n_envs,\n",
    "        ).seed(cfg.algorithm.seed)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "65a0c547",
   "metadata": {},
   "source": [
    "`iter_episodes` and `iter_partial_episodes` (autoreset) allow\n",
    "to iterate over the train workspace by sampling"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d5fd6f19",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def iter_episodes(algo: EpisodicAlgo):\n",
    "    pbar = tqdm(range(algo.cfg.algorithm.max_epochs))\n",
    "\n",
    "    train_workspace = Workspace()\n",
    "\n",
    "    for algo.epoch in pbar:\n",
    "        # Collect samples\n",
    "        train_workspace = Workspace()\n",
    "        algo.train_agent(train_workspace, t=0, stop_variable=\"env/done\")\n",
    "\n",
    "        # Update the number of steps\n",
    "        algo.nb_steps += int((~train_workspace[\"env/done\"]).sum())\n",
    "\n",
    "        # Perform a learning step\n",
    "        yield train_workspace\n",
    "\n",
    "        # Eval\n",
    "        pbar.set_description(f\"nb_steps: {algo.nb_steps}, best reward: {algo.best_reward:.2f}\")\n",
    "\n",
    "\n",
    "def iter_partial_episodes(algo: EpisodicAlgo, episode_steps: int):\n",
    "    pbar = tqdm(range(algo.cfg.algorithm.max_epochs))\n",
    "    train_workspace = Workspace()\n",
    "\n",
    "    for algo.epoch in pbar:\n",
    "        if algo.epoch > 0:\n",
    "            train_workspace.zero_grad()\n",
    "            train_workspace.copy_n_last_steps(1)\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=1, n_steps=episode_steps-1, stochastic=True\n",
    "            )\n",
    "        else:\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=0, n_steps=episode_steps, stochastic=True\n",
    "            )\n",
    "\n",
    "        algo.nb_steps += int((~train_workspace[\"env/done\"]).sum())\n",
    "        yield train_workspace\n",
    "\n",
    "        pbar.set_description(f\"nb_steps: {algo.nb_steps}, best reward: {algo.best_reward:.2f}\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "727322f3",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class EpisodicDQN(EpisodicAlgo):\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "            \n",
    "        # Get the observation / action state space dimensions\n",
    "        obs_size, act_size = self.train_env.get_obs_and_actions_sizes()\n",
    "\n",
    "        # Our discrete Q-Agent\n",
    "        self.q_agent = DiscreteQAgent(obs_size, cfg.algorithm.architecture.hidden_size, act_size)\n",
    "\n",
    "        # The e-greedy strategy (when training)\n",
    "        explorer = EGreedyActionSelector(cfg.algorithm.epsilon)\n",
    "\n",
    "        # The training agent combines the Q agent\n",
    "        self.train_policy = Agents(self.q_agent, explorer)\n",
    "\n",
    "        # The optimizer for the Q-Agent parameters\n",
    "        self.optimizer = setup_optimizer(self.cfg.optimizer, self.q_agent)\n",
    "\n",
    "        # ...and the evaluation policy (select the most likely action)\n",
    "        self.eval_policy = Agents(self.q_agent, ArgmaxActionSelector())\n",
    "\n",
    "    def run(self):\n",
    "        for train_workspace in iter_episodes(self):\n",
    "            q_values, terminated, reward, action = train_workspace[\n",
    "                \"q_values\", \"env/terminated\", \"env/reward\", \"action\"\n",
    "            ]\n",
    "        \n",
    "            # Determines whether values of the critic should be propagated\n",
    "            # True if the episode reached a time limit or if the task was not done\n",
    "            # See https://github.com/osigaud/bbrl/blob/master/docs/time_limits.md\n",
    "            must_bootstrap = ~terminated\n",
    "        \n",
    "            # Compute critic loss\n",
    "            critic_loss = compute_critic_loss(self.cfg, reward, must_bootstrap, q_values, action)\n",
    "\n",
    "            # Store the loss for tensorboard display\n",
    "            self.logger.add_log(\"critic_loss\", critic_loss, self.nb_steps)\n",
    "\n",
    "            # Gradient step\n",
    "            self.optimizer.zero_grad()\n",
    "            critic_loss.backward()\n",
    "            torch.nn.utils.clip_grad_norm_(\n",
    "                self.q_agent.parameters(), self.cfg.algorithm.max_grad_norm\n",
    "            )\n",
    "            self.optimizer.step()\n",
    "\n",
    "            # Evaluate the current policy (if needed)\n",
    "            self.evaluate()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "66e83197",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "# We setup tensorboard before running DQN\n",
    "setup_tensorboard(\"./outputs/tblogs\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "97c2f23d",
   "metadata": {},
   "outputs": [],
   "source": [
    "params={\n",
    "  \"save_best\": False,\n",
    "  \"base_dir\": \"./outputs/${gym_env.env_name}/dqn-simple-S${algorithm.seed}_${current_time:}\",\n",
    "  \"collect_stats\": True,\n",
    "  \"logger\": {\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"log_dir\": \"./outputs/tblogs/${gym_env.env_name}/dqn-simple-S${algorithm.seed}_${current_time:}\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 3,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"epsilon\": 0.02,\n",
    "    \"n_envs\": 8,\n",
    "    \"eval_interval\": 2000,\n",
    "    \"max_epochs\": 500,\n",
    "    \"nb_evals\": 10,\n",
    "    \"discount_factor\": 0.99,\n",
    "    \"architecture\":{\"hidden_size\": [128, 128]},\n",
    "  },\n",
    "  \"gym_env\":{\n",
    "    \"env_name\": \"CartPole-v1\",\n",
    "  },\n",
    "  \"optimizer\":\n",
    "  {\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 2e-3,\n",
    "  }\n",
    "}\n",
    "\n",
    "dqn = EpisodicDQN(OmegaConf.create(params))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3450054b",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Run and visualize the best agent\n",
    "dqn.run()\n",
    "dqn.visualize_best()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3eea06a5",
   "metadata": {},
   "source": [
    "## What's next?\n",
    "\n",
    "To get a full DQN, we need to do the following:\n",
    "- Add a replay buffer. We can add a replay buffer independently from the\n",
    "  target network. The version with a replay buffer and no target network\n",
    "  corresponds to [the NQF\n",
    "  algorithm](https://link.springer.com/content/pdf/10.1007/11564096_32.pdf).\n",
    "  This will be the aim of the next notebook.\n",
    "- Before adding the replay buffer, we will first move to a version of DQN\n",
    "  which uses the AutoResetGymAgent. This will be the aim of the next notebook\n",
    "  too.\n",
    "- We should also add a few extra-mechanisms which are present in the full DQN\n",
    "  version: starting to learn once the replay buffer is full enough, decreasing\n",
    "  the exploration rate epsilon...\n",
    "<!-- - We could also add visualization tools to visualize the learned Q network, by using the `plot_critic` function available in [`bbrl.visu.visu_critics`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/visu/visu_critics.py#L13) -->"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_markers": "\"\"\""
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
